#class ASN1Struct
package ASN1Struct;

use strict;
use SequenceContent;
use ChoiceContent;


#constructor
sub new {
    my $invocant = shift; #the first arg is the class
    my $class = ref($invocant) || $invocant;
    my $self = {
        explicit   => undef,  # boolean - true if it's a top-level struct (i.e., a named IE)
        specname   => undef,  # name of struct in spec
        uniquename => undef,  # unique name assigned by IEreader (same as specname for explicit structures)
        type       => undef,  # ASN1 type
        contents   => undef,  # depends on what "type" is
        precomment => "",    # comment that precedes the struct in the spec
        @_,
    };
    bless $self, $class;
    return $self;
}

# =========
# accessors
# =========
sub explicit   { return $_[0]->{explicit};   } 
sub specname   { return $_[0]->{specname};   } 
sub uniquename { return $_[0]->{uniquename}; } 
sub type       { return $_[0]->{type};       } 
sub contents   { return $_[0]->{contents};   } 
sub precomment { return $_[0]->{precomment}; } 



sub debugprint {
    my ($ob) = @_;

    my $arrow = "-->";
    if($ob->explicit)
    {
        $arrow = "==>";
    }

    print $ob->precomment;
    print $arrow." ".$ob->uniquename." ".$ob->type." ::= ";
    
    $_ = $ob->type;
    SWITCH: {
       
        /^SEQUENCE$/     && do {
            print $ob->{contents}->getNumOpts." optionals out of ".$ob->{contents}->getNumChildren." children\n";
            foreach(0..($ob->{contents}->getNumChildren-1))
            {
                my ($childname,$childuname,$type,$opt,$com) = @{$ob->{contents}->getChildArray($_)};
                
                if($com)
                {
                    print "$com";
                }

                my $asterisk = $opt ? "o" : "*";
                print "    $asterisk $childname ($childuname) $type ";
                $opt && print "OPTIONAL";
                print "\n";
            }
            last SWITCH;
        };

        /^CHOICE$/     && do {
            print $ob->{contents}->getNumChoices." options\n";
            foreach(0..($ob->{contents}->getNumChoices-1))
            {
                my ($childname,$childuname,$type) = @{$ob->{contents}->getChoiceArray($_)};
                print "    * $childname ($childuname) $type\n";
            }
            last SWITCH;
        };
        
        (/^INTEGER$/     ||
        /^OCTET STRING$/  ||
        /^BIT STRING$/)    && do {
            my @foo = @{$ob->contents};
            foreach( 0..$#{$ob->contents} )
            {
                print $ob->contents->[$_]." ";
            }
            print "\n";
            last SWITCH;
        };
       
        #default
        print "not coded to handle type '$_' yet.\n";
    }

    print "\n";
}




1;
