package HTMLprinter;

use strict;

use ASN1Struct;
use IEreader;
use DefinitionContent;
use SequenceContent;
use ChoiceContent;
use BooleanTF;

#===============
# settings you might want to mess with
#===============

# What precedes an optional element in a sequence.
# This should only be two HTML-printed characters.
my $_preOpt = '<span class="OPT">*</span> ';

# How much to indent
my $_indentStep = 5;

# Max line width
my $_maxLineWidth = 120;

#===============
# public
#===============
sub print($$);

#===============
# private
#===============

# don't mess with this
my $_indentLevel = 0;

# simple tool functions
sub printHeader($);
sub strDiv($$);
sub strSpan($$);
sub strIndent();
sub strWordWrappedList(@);

# pretty-printing functions
sub printExplicitIE($);
sub ppTypeDispatcher($;$);
sub ppInteger($);
sub ppSequence($);
sub ppString($$);  # Bit or Octet string
sub ppConstantInteger($);
sub ppEnum($);
sub ppChoice($);
sub ppSeqof($);

#values to be copied from IEreader in "print"
my @defs;
my @ies;
my %iehash; #iename -> struct
my $numex;


#=====================================================
# simple tool functions
#=====================================================
sub printHeader($)
{
    print <<HTMLHEADER;
<?xml version="1.0" encoding="utf-8" ?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">



<head>
<meta name="Author" content="Generated by ASN1 Parserator" />
<title>ASN1 HTML Generated by ASN1 Parserator</title>
<!--<link rel="stylesheet" href="asn1-style.css" type="text/css" />-->
<style type="text/css">
HTMLHEADER

    open CSSSTYLE, "<asn1-style.css";
    while(<CSSSTYLE>) { print; }
    close CSSSTYLE;

    print <<HTMLHEADER;
    
</style>

<!-- IE6 WORKAROUNDS FOR SIDEBAR MENU -->
<!-- It is impossible to get IE6 to show this page like it's supposed to be.
     Every hack or workaround creates some other wacky side-effect.
     Setting the sidebar menu in the top-right is the best that can be done.
     MS IE devs can go straight to hell.
-->
<!--[if lt IE 7.]>
<style type="text/css">
div.sidebar { position: absolute; }
</style>
<![endif]-->

</head>

<body>

<div class="sidebar">
HTMLHEADER

    my %indexItems = (
        "BCCH-BCH"    => "BCCH-BCH-Message",   
        "BCCH-FACH"   => "BCCH-FACH-Message",
        "CCCH DL"     => "DL-CCCH-Message",
        "CCCH UL"     => "UL-CCCH-Message",
        "DCCH DL"     => "DL-DCCH-Message",
        "DCCH UL"     => "UL-DCCH-Message",
        "HO To UTRAN" => "HandoverToUTRANCommand",
        "InterRAT HO" => "InterRATHandoverInfo",
        "MCCH"        => "MCCH-Message",
        "MSCH"        => "MSCH-Message",
        "PCCH"        => "PCCH-Message",
        "SHCCH DL"    => "DL-SHCCH-Message",
        "UE RAC Info" => "UE-RadioAccessCapabilityInfo"
    );

    print "<br />\n<strong>INDEX:</strong><br />\n";
    
    foreach( sort keys %indexItems )
    {
        my $val = $indexItems{$_};
        if( exists $iehash{$val} )
        {
            print "<a href=\"#$val\">$_</a><br />\n";
        }
    }

    print "<br />\n<strong>SIB IEs:</strong><br />\n";
    
    if( exists $iehash{"MasterInformationBlock"} )
    { print "<a href=\"#MasterInformationBlock\">MIB</a><br />\n"; }

    print "<pre>";
    foreach( 1..18 )
    {
        my $sib = "SysInfoType$_";
        if($_ == 5) { $sib .= "bis"; }
        
        if( exists $iehash{$sib} ) 
        {
            my $w = ($_ < 10) ? " " : "";
            my $n = ( ($_==4) || ($_==8) || ($_==12) || ($_==16) )
                    ? "\n" : "";
            print "<a href=\"#$sib\">$w$_</a> $n";
        }
    }
    print "</pre>";
  
    if( exists $iehash{"SysInfoTypeSB1"} )
    { print "<a href=\"#SysInfoTypeSB1\">SB1</a>\n"; }
    if( exists $iehash{"SysInfoTypeSB2"} )
    { print "<a href=\"#SysInfoTypeSB2\">SB2</a><br />\n"; }

#This sidebar will not stay fixed in place with Internet Explorer 6.

    print <<IENOTICE;
<!--[if lt IE 7.]>
<div class="ienotice" style="width: auto">
In browsers other than IE6, this sidebar menu
will be stationary.<br />
This is an IE6 bug.
</div>
<![endif]-->
IENOTICE


    my $asn1filename = $_[0];
    print <<HTMLHEADER;
</div>

<div class="main">

<div class="headerinfo">
<h1>Generated by ASN1 Parserator v1.1</h1>
The handy ASN1-to-HTML generator tool.<br />
<a href="http://grantb.net/staticpages/parserator">http://grantb.net/parserator</a><br />
Written by Grant Birchmeier. 
<br /><br />
Generated from input file: $asn1filename
</div>

HTMLHEADER

}


#  (arg0) class  (arg1) string
sub strDiv($$)
{
    if($_[0] eq "COMMENT") { chomp($_[1]); }

    return "<div class=\"$_[0]\">$_[1]</div><!--$_[0]-->\n";
}


#  (arg0) class  (arg1) string
sub strSpan($$)
{
    # if COMMENT, indent every line
    if($_[0] eq "COMMENT")
    {
        chomp($_[1]);

        my @lines = split /\n/,$_[1];
        $_[1] = "";
        foreach(@lines) { $_[1] .= strIndent().$_."\n"; }
        chomp $_[1];
    }

    return "<span class=\"$_[0]\">$_[1]</span>";
}


sub strIndent()
{
    return " " x ($_indentLevel*$_indentStep);
}

sub strWordWrappedList(@)
{
    my @a = @_;

    my $line = strIndent().shift(@a);
    my $str;
    
    while(@a)
    {
        my $w = shift @a;
        if( (length($line)+length($w)) <= $_maxLineWidth )
        {
            $line .= $w;
        }
        else
        {
            $str .= "$line\n";
            $line = strIndent().$w;
        }
    }
    $str .= "$line\n";

    return $str;
}


#=====================================================
# Main printing function
# (goes through each definition and all IEs inside them)
# arg0=html_filename   arg1=asn1_filename
#=====================================================
sub print($$)
{
    open( HTMLFILE, ">", "$_[0]" ) or die "Can't open html file";
    select( HTMLFILE );

    #=========================================================
    # sloppily convert the class vars to useful arrays/hashes
    my $vscalar = IEreader::getVars();
    my @v = @{$vscalar};

    @defs = @{$v[0]};
    @ies  = @{$v[1]};
    %iehash = @{$v[2]};
    $numex = $v[3];
    #=========================================================

#    print 'size of @defs: '.scalar(@defs)."\n";
#    print 'size of @ies: '.scalar(@ies)."\n";
#    print 'size of %hash: '.scalar(keys %iehash)."\n";
#    print '$numex: '."$numex\n";

#    my @keyarray = sort keys %iehash;
#    my $kstr = $keyarray[15];
#    my $kval = $iehash{$kstr};
#    my $ob = $ies[$kval];

#    print "key '$kstr' has val $kval:  ".$ob->uniquename."\n";
    printHeader($_[1]);

    foreach( @defs )
    {
        print '<div class="DEFINITION">'."\n";
        if($_->{pre_def_comment})
        {
            print "<pre>".strSpan("COMMENT",$_->{pre_def_comment})."</pre>\n";
        }
        print strDiv("DECL",$_->{name}." DEFINITIONS AUTOMATIC TAGS ::=");
        
        if($_->{pre_begin_comment})
        {
            print "<pre>".strSpan("COMMENT",$_->{pre_begin_comment})."</pre>\n";
        }
        print strDiv("BEGIN_END","BEGIN");
        
        if($_->{pre_import_comment})
        {
            print "<pre>".strSpan("COMMENT",$_->{pre_import_comment})."</pre>\n";
        }

        #TODO what if the input has no imports sections?  Fix it so it's not printed!
        print strDiv("IMPORTS","IMPORTS");

        print '<div class="IMPORT_SECTION">'."\n";
        for( my $i=0; $i< $_->getNumFroms(); $i++ )
        {
            my @ies = $_->getImportedIEs($i);

            print '<div class="IMPORTED_IES">'."\n";
            foreach my $x (@ies)
            {
                # check for comment
                my $comment = $_->getImpIEcomment($x);
                if($comment)
                {
                    print "<pre>".strSpan("COMMENT",$comment)."</pre>\n";
                }

                # determine if should end in comma or newline
                # (if this is the last IE, then newline)
                my $end = ($x eq $ies[$#ies]) ? "\n" : ", ";
                print "$x$end";
            }
            print "</div><!--IMPORTED_IES-->\n";
             
            my $semicolon = ($i==$_->getNumFroms()-1) ? ";" : "";
            print strDiv("FROM","FROM ".$_->getFromName($i).$semicolon);
        }
        print '</div><!--IMPORT_SECTION-->'."\n\n";

        my @defIEs = $_->getIEs();
        print "<pre>\n";
        foreach(@defIEs)
        {
            printExplicitIE($_);
            print "\n";
        }
        print "</pre>\n";


        print strDiv("BEGIN_END",
                     "END ".strSpan("COMMENT","--******** ".$_->{name}." ********"));
        print "</div><!--DEFINITION: ".$_->{name}."-->\n\n\n"; 
    }

    print "</div><!--main-->\n\n</body>\n</html>\n";
    select(STDOUT);
}



#=====================================================
# Pretty-printing functions
#=====================================================

# Input is IE name.
sub printExplicitIE($)
{
    my $ie = $ies[$iehash{$_[0]}];

    my ($name, $type) = ($ie->specname, $ie->type);
    print "<a name=\"$name\"></a>";
    if ($ie->precomment) { print strSpan("COMMENT",$ie->precomment)."\n"; } 

    if ($type eq "CONSTANT_INTEGER")
    {
        print "$name <span class=\"TERMTYPE\">INTEGER</span> ::= ";
    }
    else
    {
        print "$name ::= ";
    }

    print ppTypeDispatcher($_[0]);
       
    print "\n";

}

# Input is IE uname.  Second optional input is type (only used for non-core types).
sub ppTypeDispatcher($;$)
{
    my $ie = exists($iehash{$_[0]}) ? $ies[$iehash{$_[0]}] : undef;
    my $noncoretype = $_[1];

    # Defined non-core type; just return what was given
    if( !$ie )
    {
        return "<a href=\"#$_[1]\">$_[1]</a>";
    }

    # ASN1 core types
    if   ($ie->type eq "INTEGER")      { return ppInteger    ($ie->contents); }
    elsif($ie->type eq "BIT STRING")   { return ppString     ("BIT",  $ie->contents); }
    elsif($ie->type eq "OCTET STRING") { return ppString     ("OCTET",$ie->contents); }
    elsif($ie->type eq "ENUMERATED")   { return ppEnum       ($ie->contents); }
    elsif($ie->type eq "SEQUENCE")     { return ppSequence   ($ie->contents); }
    elsif($ie->type eq "CHOICE")       { return ppChoice     ($ie->contents); }
    elsif($ie->type eq "BOOLEAN")      { return "<span class=\"TERMTYPE\">BOOLEAN</span>"; }
    elsif($ie->type eq "NULL")         { return "<span class=\"TERMTYPE\">NULL</span>"; }

    # This quasi-type has been convenience-ized
    elsif($ie->type eq "SEQ-OF")       { return ppSeqof      ($ie->contents); }
    
    # Non-core types; created by IEReader for parserator purposes
    elsif($ie->type eq "ALIAS")        { return '<a href="#'.$ie->contents.'">'.$ie->contents.'</a>'; }
    elsif($ie->type eq "CONSTANT_INTEGER")  { return ppConstantInteger($ie->contents); }


    print STDERR "! Non-core type not supported yet: @_\n";
}


sub ppInteger($)
{
    my @contents = @{$_[0]};
    if( scalar(@contents)==0 )
    {
        return "<span class=\"TERMTYPE\">INTEGER</span>";
    }

    my ($lb,$ub) = @contents;

    # turn any constant lb/ub into hyperlinks
    if( $lb !~ /^-?[0-9]*$/ )  { $lb = "<a href=\"#$lb\">$lb</a>"; }
    if( $ub !~ /^-?[0-9]*$/ )  { $ub = "<a href=\"#$ub\">$ub</a>"; }

    return "<span class=\"TERMTYPE\">INTEGER</span> <span class=\"TYPE_AUX\">($lb..$ub)</span>";
}


# arg1={"BIT"|"OCTET"}, arg2=arrayref: lb/ub, or -1 if unbounded
sub ppString($$)
{
    my ($type,$cref) = @_;

    my $ret = "<span class=\"TERMTYPE\">$type STRING</span>";

    # Handle unbounded
    if( $cref == -1 ) { return $ret; }

    # Handle containing
    if( $cref->[0] eq "--CONTAINING" )
    {
        my $t = $cref->[1];
        return "$ret<span class=\"TYPE_AUX\"> (CONTAINING <a href=\"#$t\">$t</a>)</span>";
    }

    # Handle bit field type
    # TODO check with spec to see if "aa(0), bb(1), cc(2)" is the only way
    if( $cref->[0] eq "--BITFIELD" )
    {
        $ret .= " {\n";

        # fc = field count
        my $fc_index = 1;

        $_indentLevel++;
        if( $cref->[1] =~ /^--/ )
        {
            $ret .= strSpan("COMMENT",$cref->[1])."\n";
            $fc_index = 2;
        }

        my @items;
        for( my $i=0; $i<$cref->[$fc_index]; $i++ )
        {
            my $str = $cref->[1+$fc_index+$i]."($i)";
            if( $i != $cref->[$fc_index]-1 ) { $str .= ", "; }

            push @items, $str;
        }
        $ret .= strWordWrappedList(@items);

        $_indentLevel--;

        $ret .= strIndent()."} <span class=\"TYPE_AUX\">(SIZE(".$cref->[$fc_index]."))</span>";
        return $ret;
    }

    # Handle regular bounded-length strings
    # turn any constant lb/ub into hyperlinks
    my ($lb,$ub) = @{$cref};
    if( $lb !~ /^-?[0-9]*$/ )  { $lb = "<a href=\"#$lb\">$lb</a>"; }
    if( $ub !~ /^-?[0-9]*$/ )  { $ub = "<a href=\"#$ub\">$ub</a>"; }
    
    if( $lb eq $ub ) {
        return "$ret<span class=\"TYPE_AUX\"> (SIZE($lb))</span>";
    }else{
        return "$ret<span class=\"TYPE_AUX\"> (SIZE($lb..$ub))</span>";
    }
}


# TODO line them up nicer
sub ppConstantInteger($)
{
    my ($c,$eolc) = @{$_[0]};

    # TODO turn any constant lb/ub into hyperlinks
    # sanity check, prints warning if a non-integer constant was found
    if( $c !~ /^-?[0-9]*$/ )  { print STDERR "=======Non-integer constant?\n"; }

    # print end-of-line comment, if any
    if( $eolc )
    {
        $eolc = "  ".strSpan("COMMENT",$eolc);
    }
    else
    {
        $eolc = "";
    }
    

    return "$c$eolc";
}


# TODO - do wrap behavior based on line length instead of #elements
# TODO - might need to special case ReportingIntervalLong comment
sub ppEnum($)
{
    my @a = @{$_[0]};

    my $retstring = "<span class=\"TERMTYPE\">ENUMERATED</span> {";

    # Comments (if any) are mixed in the array with the enum items
    # Find if comments are present, and find last array index with a non-comment
    my $comments_present = FALSE;
    my $last_nc_index = -1;
    for( my $i=0; $i<=$#a; $i++ )
    {
        if( $a[$i] !~ /^--/ )
        {
            $last_nc_index = $i;
        }
        else
        {
            $comments_present = TRUE;
        }
    }

    if($comments_present)
    {
        # TODO should items be newlined?  Investigate later.
        $retstring .= "\n";
        $_indentLevel++;
        for( my $i=0; $i<=$#a; $i++ )
        {
            if( $a[$i] =~ /^--/ )
            {
                $retstring .= strSpan("COMMENT",$a[$i])."\n";
            }
            else
            {
                $retstring .= strIndent().$a[$i];
                if( $i<$last_nc_index )
                {
                    $retstring .= ",";
                }
                $retstring .= "\n";
            }
        }
        $_indentLevel--;
        $retstring .= strIndent();
    }
    else
    {
        #short enum
        if( (length(strIndent()."@a")+30+scalar(@a)) < $_maxLineWidth )
        {
            foreach(@a) { $retstring .= "$_, "; }
            chop $retstring; chop $retstring;
        }
        #long enum
        else
        {
            $retstring .= "\n";
            $_indentLevel++;
            foreach(@a)
            {
                $retstring .= strIndent()."$_,\n";
            }
            $_indentLevel--;
            chop $retstring; chop $retstring;
            $retstring .= "\n".strIndent();
        }
    }

    $retstring .= "}";

    return $retstring;
}


#TODO if only one child, no newlines (UL-UM-RLC-Mode) - maybe base wrapping on string length instead
sub ppSequence($)
{
    my $seqcont = $_[0];

    my $numChildren = $seqcont->getNumChildren();

    # If no children, stop now.  No more processing needed.
    if($numChildren==0)
    {
        return "<span class=\"TYPE\">SEQUENCE</span> {}";
    }

    my $retstring = "<span class=\"TYPE\">SEQUENCE</span> {\n";
    my $nameFieldWidth = $seqcont->getMaxNameWidth();

    # for each child, append it to the string
    $_indentLevel++;
    for( my $i=0; $i<$numChildren; $i++ )
    {
        my @c = @{$seqcont->getChildArray($i)};  # @c = [name,uname,type,opt,preceding comment,isdef,def,eol comment]

        my $childstr = "";

        # suffix all but the last one with a comma
        my $comma = (($i<$numChildren-1) || $seqcont->getHasDotDotDot()) ? "," : "";

        # check if it's optional or default (only one should be true)
        my $opt   = ($c[3]) ? " <span class=\"OPTIONAL\">OPTIONAL</span>" : "";
        my $def   = ($c[5]) ? " <span class=\"OPTIONAL\">DEFAULT</span> $c[6]" : "";
        if($c[3] && $c[5]) { #sanity
            die("Logical error in ppSequenceChildren!  Somehow a seq child is both optional _and_ default!");
        }

        $childstr = ppTypeDispatcher( $c[1],$c[2] );

        # pad the child names all but sequences and choices
        # TODO this ain't right; seq/choices influence name width, but are not affected by it!
        if( ($c[2] eq "CHOICE")
            || (($c[2] eq "SEQUENCE") && ($ies[$iehash{$c[1]}]->contents->getNumChildren() > 0)) )
        {
            $childstr = "$c[0] $childstr";
        }
        else
        {
            my $paddedName = sprintf("%-${nameFieldWidth}s", $c[0]);
            $childstr = "$paddedName $childstr";
        }
        
        # indent
        # also add a thing to precede each optional
        my $indent = strIndent();
        if( $c[3] ) { chop $indent; chop $indent; $indent .= $_preOpt; }
        $childstr = $indent.$childstr;

        # add preceeding comment, if any
        # TODO figure out why opening newlines are knocked.
        if($c[4]) { $childstr = strSpan("COMMENT",$c[4]) . "\n$childstr"; }

        my $eolcomment = "";
        if( $c[7] )
        {
            $eolcomment = "     ".$c[7];
        }

        $retstring .= "$childstr$opt$def$comma$eolcomment\n";
    } #for each choice

    if($seqcont->getHasDotDotDot())
    {
        $retstring .= strIndent()."...\n";
    }

    my $tc = $seqcont->getTailComment();
    if($tc)
    {
        $retstring .= strSpan("COMMENT",$tc)."\n";
    }

    $_indentLevel--;
    
    $retstring .= strIndent()."}";
    return $retstring;
}


sub ppChoice($)
{
    my $chcont = $_[0];

    my $numChoices = $chcont->getNumChoices();
    
    my $retstring = "<span class=\"TYPE\">CHOICE</span> {\n";
    my $nameFieldWidth = $chcont->getMaxNameWidth();

    # for each child, append it to the string
    $_indentLevel++;
    for( my $i=0; $i<$numChoices; $i++ )
    {
        my @c = @{$chcont->getChoiceArray($i)};  # @c = [name,uname,type,preceding comment,eol comment]

        my $childstr = "";

        # suffix all but the last one with a comma
        my $comma = (($i<$numChoices-1) || $chcont->getHasDotDotDot()) ? "," : "";

        $childstr = ppTypeDispatcher( $c[1],$c[2] );

        # pad the child names all but sequences and choices
        # TODO this ain't right; seq/choices influence name width, but are not affected by it!
        if( ($c[2] eq "CHOICE")
            || (($c[2] eq "SEQUENCE") && ($ies[$iehash{$c[1]}]->contents->getNumChildren() > 0)) )
        {
            $childstr = "$c[0] $childstr";
        }
        else
        {
            my $paddedName = sprintf("%-${nameFieldWidth}s", $c[0]);
            $childstr = "$paddedName $childstr";
        }
        
        # indent
        $childstr = strIndent().$childstr.$comma;

        # add preceeding and eol comments, if any
        if($c[3]) { $childstr = strSpan("COMMENT",$c[3]) . "\n$childstr"; }
        if($c[4]) { $childstr .= "    ".strSpan("COMMENT",$c[4]); }

        $retstring .= "$childstr\n";
    } #for each choice

    if($chcont->getHasDotDotDot())
    {
        $retstring .= strIndent()."...\n";
    }

    my $tc = $chcont->getTailComment();
    if($tc)
    {
        $retstring .= strSpan("COMMENT",$tc)."\n";
    }

    $_indentLevel--;

    $retstring .= strIndent()."}";
    return $retstring;
}


# TODO newlines aren't right for TFCS-ReconfAdd (seq child of seq-of)
sub ppSeqof($)
{
    my ($uname,$lb,$ub) = @{$_[0]};

    # Check im/explicit status here.  We'll need it later.
    # TODO this is a questionable assignment.  What the hell was I doing here?  Shouldn't it be "= exists && x->explicit" ?  Why the second test anyway?
    my $explicit = !exists($iehash{$uname}) || $ies[$iehash{$uname}]->explicit;

    # turn any constant lb/ub into hyperlinks
    if( $lb !~ /^[0-9]*$/ )  { $lb = "<a href=\"#$lb\">$lb</a>"; }
    if( $ub !~ /^[0-9]*$/ )  { $ub = "<a href=\"#$ub\">$ub</a>"; }

    # construct the SIZE(xxx) substring
    my $sizestr = '<span class="TYPE_AUX">(SIZE(';
    if($lb eq $ub) { $sizestr .= "$lb";      }
    else           { $sizestr .= "$lb..$ub"; }
    $sizestr .= "))</span>";

    # note the post-"OF" comment if there is one (it will be arg[3])
    # (an example is RF-CapabBandListFDDComp)
    my $of_comment = "";
    if( $#{$_[0]} == 3 )
    {
        $of_comment = strSpan("COMMENT",$_[0]->[3]);
    }

    # Finish it up
    if($explicit) # no line wraps for explicit (unless there's a comment)
    {

        if( $of_comment )
        {
            $_indentLevel++;
            # pass a nonsense string that can't possibly match any other IE
            # (basically we're just using ppTypeDispatcher to html-ize the type)
            my $type = ppTypeDispatcher("Adam West is Batman",$uname);
            my $s = "<span class=\"TYPE\">SEQUENCE $sizestr OF</span>\n"
                   .strIndent().$of_comment."\n"
                   .strIndent().$type;
            $_indentLevel--;
            
            return $s;
        }
        else
        {
            # pass a nonsense string that can't possibly match any other IE
            # (basically we're just using ppTypeDispatcher to html-ize the type)
            my $type = ppTypeDispatcher("Adam West is Batman",$uname);
            return "<span class=\"TYPE\">SEQUENCE $sizestr OF</span> $type";
        }
    }
    else
    {
        #TODO might not want to wrap a line if not a SEQ or CHOICE
        $_indentLevel++;
        my $type = ppTypeDispatcher($uname);
        my $s = "<span class=\"TYPE\">SEQUENCE $sizestr OF</span>\n";
        if($of_comment) { $s .= strIndent().$of_comment."\n"; }
        $s .= strIndent().$type;
        $_indentLevel--;

        return $s;
    }

}















1;
