package IEreader;

use strict;
use BooleanTF;
use ASN1tokenizer;
use ASN1Struct;
use ASN1Token;
use SequenceContent;
use ChoiceContent;
use DefinitionContent;


#------------------------------------------------------------------------------
# "PUBLIC" PROTOTYPES
# Driver programs should use these functions and nothing else from this file.
# They are not exported, so call them like "IEreader::function()".
#------------------------------------------------------------------------------
sub initFromStdin();

sub debugtestprint();

#==============================================
# Everything below here is "private".
# Users of this module should only use the interface above.
#==============================================

#----------------------
# vars
#----------------------
my @_defArray;   # array of DefinitionContents

my @_iesArray;   # array of all ASN1Structs
my %_iesByName;  # key=IEname (uniquename), value=indexPlacementInArray

my $_numExplicitIEs=0; # count how many declared IEs
                       # (i.e. those that aren't implicitly defined inside another struct)

sub getVars() { return [ [@_defArray],
                         [@_iesArray],
                         [%_iesByName],
                         $_numExplicitIEs ] };

#----------------------
# private prototypes
#----------------------

# utils
sub grabToken(); # skips EOL_COMMENT tokens, prints a note if they were found
                 # (doesn't skip BLOCK_COMMENTS)
sub ncpeek(;$); # like PEEK, but skips comments
sub getNoncommentToken();  # skips comment tokens, prints a note if they were found
sub matchOrFail($;$);  # match the input string or die (2nd param is for debug info);
                       # (skips comments with getNoncommentToken())
sub isAsn1CoreType($);  # test if string matches SEQUENCE, CHOICE, etc

# structure type processing functions
sub processDefinition();
sub processExplicitStruct();    # arg is optional prefix name;
sub processImplicitStruct($$$); # uniquename, specname, type
sub processSequence($);  # arg is prefix name (might be "")
sub processChoice($);    # arg is prefix name (might be "")
sub processSeqOf($);     # arg is prefix name (might be "")
sub processInteger();
sub processBitOrOctetString();
sub processSize();
sub processBoolean();
sub processNull();
sub processEnum();
sub processConstantInteger();



#----------------------
# function definitions
#----------------------

sub initFromStdin()
{
    ASN1tokenizer::initFromStdin();

    print "==============================================\n";
    print " IE READER - building internal representation\n"
         ."             of ASN1 structures\n";
    print "==============================================\n";
    
#print "   ".ASN1tokenizer::peek(1)->string."\n";
#print "   ".ASN1tokenizer::peek(2)->string."\n";
#print "   ".ASN1tokenizer::peek(3)->string."\n";
#print "   ".ASN1tokenizer::peek(4)->string."\n";
#print "   ".ASN1tokenizer::peek(5)->string."\n";
#die();


    while(ASN1tokenizer::peek(1) && (ASN1tokenizer::peek(1)->string ne "SUPERDIE!"))
    {
        processDefinition();
    }



    #TODO distinguish between EOF and SUPERDIE!
#    print "died because:\n";
#    print "   ".ASN1tokenizer::peek(1)."\n";
#    print "   ".ASN1tokenizer::peek(1)->string."\n";

    print " Done! (".scalar(@_iesArray)
         ." IEs: $_numExplicitIEs declared, ";
    print ( scalar(@_iesArray) - $_numExplicitIEs );
    print " implicit.)\n\n";
}


sub debugtestprint()
{
    print "size of array: ".scalar(@_iesArray)." / size of hash: ".scalar(keys(%_iesByName))."\n\n";

    if(FALSE) {  #lazy way to disable this block
    foreach my $def (@_defArray)
    {
        bless ($def, "DefinitionContent");

        print "name: $def->{name}\n";
        print "pre_def:\n $def->{pre_def_comment}\n";
        print "pre_begin:\n $def->{pre_begin_comment}\n";
        print "pre_import:\n $def->{pre_import_comment}\n";

        foreach( my $i=0; $i<$#{$def->{from_array}}; $i++ )
        {
            print "imported from ${$def->{from_array}}[$i]\n";
            print "   last element is ${$def->{from_end_indices}}[$i] ";
            print ${$def->{imported_ie_array}}[ ${$def->{from_end_indices}}[$i] ]."\n";
        }

        print "last_index: $def->{index_of_last_struct_in_def} ";
        print $_iesArray[$def->{index_of_last_struct_in_def}]->uniquename ."\n";
        
        print "===========================================\n";
    }
    }#if(FALSE)


    for( my $i=0; $i < scalar(@_iesArray); $i++ )
    {
        $_iesArray[$i]->debugprint();

        if( !exists $_iesByName{$_iesArray[$i]->uniquename} )  { print "!!! doesn't exist in hash !!!\n" }
    }
    
}


#========================================
# utils
#========================================

# gets a token, but ignores and flags EOL_COMMENT tokens
sub grabToken()
{
    while(my $token = ASN1tokenizer::getToken())
    {
        if( $token->type != ASN1Token::EOL_COMMENT )
        {
            return $token;
        }
        else
        {
            print "============================================================================Ignored comment token type EOL_COMMENT from source line ".$token->line.".\n"; 
        }
    }

    # no tokens left
    return FALSE;
}

#------------------------------------
# "ncpeek" = 'no-comment' peek
# Same as ASN1tokenizer::peek(), but skips over comment tokens
# (don't need to print a note; other functions will do it when the token is grabbed)
#------------------------------------
sub ncpeek(;$)
{
    my $target = 1;
    my $nc_token_count = 0;
    my $i = 1;

    if( defined($_[0]) )
    {
        if( $_[0] <= 0 )
        {
            die("ASN1tokenizer::peek(): argument must be >0");
        }
        else
        {
            $target = $_[0];
        }
    }


    while( $nc_token_count != $target )
    {
        my $foo = ASN1tokenizer::peek($i);
        
        if( $foo == FALSE )
        {
            return FALSE;
        } 
        elsif( $foo->type == ASN1Token::TOKEN )
        {
            $nc_token_count++;
        }
        #else it's a comment and do nothing

        $i++;
    }

    return ASN1tokenizer::peek($i-1);
}
    

# Returns the next non-comment token.
# Call this when you want a token, but don't expect to handle a comment token.
# This will print a note if a comment was skipped, so you can fix the program to handle it if you want.
# If no tokens left, return FALSE.
sub getNoncommentToken()
{
#my $hey = FALSE;
    while(my $token = ASN1tokenizer::getToken())
    {
        if( $token->type == ASN1Token::TOKEN )
        {
            return $token;
        }
        else
        {
            my $foo = ($token->type == ASN1Token::BLOCK_COMMENT) ? "BLOCK_COMMENT" : "EOL_COMMENT";
            print "============================================================================Ignored comment token type ".$foo." from source line ".$token->line.".\n";
        }
    }

    # no tokens left
    return FALSE;
} 

# Get the next token, which is expected to match the param.
# If it doesn't match, die.
# (uses getNoncommentToken to skip comments)
sub matchOrFail($;$)
{
    my $match = $_[0];
    
    my $token = getNoncommentToken();
    if( $token->string ne $match )
    {
        my $diestring = "Token ".$token->string." from input line "
                       .$token->line." didn't match '$match' as expected.";
        if($_[1])
        {
            $diestring .= " ($_[1]) ";
        }
        
        die( $diestring );
    }
}

# Return TRUE if the string matches one of the ASN1 core types
# (e.g. SEQUENCE, CHOICE, INTEGER, etc... )
sub isAsn1CoreType($)
{
    my $t = $_[0];

    if( ($t eq "SEQUENCE")
        || ($t eq "CHOICE")
        || ($t eq "BIT STRING")
        || ($t eq "OCTET STRING")
        || ($t eq "INTEGER")
        || ($t eq "BOOLEAN")
        || ($t eq "NULL")
        || ($t eq "ENUMERATED")
        || ($t eq "SEQ-OF")      # spec doesn't use this string (obviously)
      )
    {
        return TRUE;
    }
    
    return FALSE;
}

#========================================
# structure type processing functions
#========================================

sub processDefinition()
{
    my $def = new DefinitionContent();
    
    my $token = grabToken();
    if($token->type == ASN1Token::BLOCK_COMMENT)
    {
        $def->{pre_def_comment} .= $token->string;
        $token = grabToken();
    }

    $def->{name} = $token->string;
    matchOrFail("DEFINITIONS");
    matchOrFail("AUTOMATIC");
    matchOrFail("TAGS");
    matchOrFail("::=");


    # Check for comment before "BEGIN"
    if( ASN1tokenizer::peek(1)->type == ASN1Token::BLOCK_COMMENT )
    {
        $def->{pre_begin_comment} .= grabToken()->string;
    }
    matchOrFail("BEGIN");


    # only do the following part if the import section is present
    if( (ASN1tokenizer::peek(1)->string eq "IMPORTS") ||
        (ASN1tokenizer::peek(2)->string eq "IMPORTS") ) 
    {
        # Check for comment before "IMPORTS"
        if(ASN1tokenizer::peek(1)->type == ASN1Token::BLOCK_COMMENT)
        {
            $def->{pre_import_comment} .= grabToken()->string;
        }
        matchOrFail("IMPORTS");


        # Read all the imported IEs
        while(1)
        {
            $token = grabToken();

            my $comment;
            if( $token->type == ASN1Token::BLOCK_COMMENT )
            {
                $comment = $token->string;
                $token = grabToken();
            }


            if($token->string eq "FROM")
            {
                if(defined($comment))
                {
                    print "=================== Lost a comment before a FROM in an imports list\n";
                }
                
                # The next token should be the FROM source.  Read it now.
                $def->addFrom(getNoncommentToken()->string);
            }
            elsif( $token->string eq ";")
            {
                if(defined($comment))
                {
                    print "=================== Lost a comment before a semicolon in an imports list\n";
                }
                
                # done with this whole section
                last;
            }
            elsif( $token->string ne ",")
            {
                # this is an IE, store it
                $def->addImportedIE($token->string,$comment);
            }
            else #comma
            {
                if(defined($comment))
                {
                    print "=================== Lost a comment before a comma in an imports list\n";
                }
            }   
        }
    }


    # At this point, $token should be the ";" at the end of the imports list.
    # That means the next tokens are the start of the ASN1 structures.
    while( !((ncpeek()->type   == ASN1Token::TOKEN) &&
             (ncpeek()->string eq "END")) )
    {
        my $iename = processExplicitStruct();
        # print $iename."\n";  #handy for debugging
        $def->addIE( $iename );
    }

    matchOrFail("END");

    #TODO if end of spec is like this:
    #      someIe INTEGER ::= 32 -- Maximum number of KeNB* and shortMAC-I forwarded
    #                            -- at handover for re-establishment preparation
    # Then that second spec line is gonna get lost.

    $def->{index_of_last_struct_in_def} = $#_iesArray;

    push @_defArray, $def;
}


#returns the name of the processed struct (name is same as uniquename)
sub processExplicitStruct()
{
  #=================================================================
  # 1. Get the name of the struct and any comment that precedes it.
  #    Then consume the ::= if this is an explicit structure.
    my $token = grabToken();

    my $headerComment = "";
    if($token->type == ASN1Token::BLOCK_COMMENT)
    {
        $headerComment = $token->string;
        $token = grabToken();
    }

    my $ieName = $token->string;
    my $nameprefix = $ieName."__";

    # handle constant type "sometype INTEGER ::= N"
    # (as opposed to       "sometype ::= INTEGER N")
    my $ieType = undef;
    if(ncpeek()->string eq "INTEGER")
    {
        matchOrFail("INTEGER");
        $ieType = "CONSTANT_INTEGER";
    }
    matchOrFail("::=");

    
  #=================================================================
  # 2. Get the type of the struct.
  #    Call the appropriate subroutine for its type to fill the contents array.
    if( !defined($ieType) )
    {
        $ieType = getNoncommentToken()->{string};
    }

    if( ($ieType eq "BIT") || ($ieType eq "OCTET") )
    {
        matchOrFail("STRING");
        $ieType .= " STRING";
    }
    elsif( ($ieType eq "SEQUENCE") && (ncpeek()->string eq "(") )
    {
        $ieType = "SEQ-OF";
    }

    my $ieContents;


    # switch ($ieType)
    if( $ieType eq "SEQUENCE" )
    {
        $ieContents = processSequence($nameprefix);
    }
    elsif( $ieType eq "CHOICE" )
    {
        $ieContents = processChoice($nameprefix);
    }
    elsif( $ieType eq "SEQ-OF" )
    {
        $ieContents = processSeqOf($nameprefix);
    }
    elsif(    ($ieType eq "BIT STRING"  ) 
           || ($ieType eq "OCTET STRING") 
         )
    {
        $ieContents = processBitOrOctetString();
    }
    elsif( $ieType eq "BOOLEAN" )
    {
        $ieContents = processBoolean();
    }
    elsif( $ieType eq "NULL" )
    {
        $ieContents = processNull();
    }
    elsif( $ieType eq "ENUMERATED" )
    {
        $ieContents = processEnum();
    }
    elsif($ieType eq "INTEGER")
    {
        $ieContents = processInteger();
    }
    elsif($ieType eq "CONSTANT_INTEGER") 
    {
        $ieContents = processConstantInteger();
    }
    else # ALIAS type?
    {
        # if type is all caps, it might not be an alias
        if( $ieType =~ /^[A-Z]$/ )
        {
            print "***Type is: $ieType\n";
            die("***Handling for this type not implemented yet.");
        }
        else
        {
            $ieContents = $ieType;
            $ieType = "ALIAS";
        }
    }

    push @_iesArray, new ASN1Struct( specname    => $ieName,
                                     uniquename  => $ieName,
                                     type        => $ieType,
                                     explicit    => TRUE,
                                     contents    => $ieContents,
                                     precomment  => $headerComment );

    $_numExplicitIEs++;

    
    if( exists $_iesByName{$ieName} )
    {
        die("More than one IE named $ieName exists!");
    } 
    else
    {
        $_iesByName{$ieName} = $#_iesArray;  #set this hash element to last index of the array
    }

    return $ieName;
}


# parameters: uniquename, specname, type
sub processImplicitStruct($$$)
{
    my ($ieUniqueName,$ieSpecName,$ieType) = @_;

    my $nameprefix = $ieUniqueName."__";
    my $ieContents;
    my $headerComment = "";  #need to redo this later


    # switch ($ieType)
    if( $ieType eq "SEQUENCE" )
    {
        $ieContents = processSequence($nameprefix);
    }
    elsif( $ieType eq "CHOICE" )
    {
        $ieContents = processChoice($nameprefix);
    }
    elsif( $ieType eq "SEQ-OF" )
    {
        $ieContents = processSeqOf($nameprefix);
    }
    elsif(    ($ieType eq "BIT STRING"  ) 
           || ($ieType eq "OCTET STRING") 
         )
    {
        $ieContents = processBitOrOctetString();
    }
    elsif( $ieType eq "BOOLEAN" )
    {
        $ieContents = processBoolean();
    }
    elsif( $ieType eq "NULL" )
    {
        $ieContents = processNull();
    }
    elsif( $ieType eq "ENUMERATED" )
    {
        $ieContents = processEnum();
    }
    elsif( $ieType eq "INTEGER" )
    {
        $ieContents = processInteger();
    }
    else
    {
        print "***Type is: $ieType\n";
        die("***Handling for this type not implemented yet.");
    }

    push @_iesArray, new ASN1Struct( specname    => $ieSpecName,
                                     uniquename  => $ieUniqueName,
                                     type        => $ieType,
                                     explicit    => FALSE,
                                     contents    => $ieContents,
                                     precomment  => $headerComment );

    if( exists $_iesByName{$ieUniqueName} )
    {
        die("More than one IE named $ieUniqueName exists!");
    } 
    else
    {
        $_iesByName{$ieUniqueName} = $#_iesArray;  #set this hash element to last index of the array
    }
}

sub processSequence($)
{
    my $nameprefix = $_[0];

    # Parsing starts on the next token after the word SEQUENCE,
    # (which we expect is going to be a "{").
    matchOrFail("{");
    
    my (@name, @uname, @ctype, @opt, @comments, @isdef, @def, @eolcomments);
    my $hasDotDotDot = FALSE;

    while(1)
    {
        # Check if it's an empty sequence.  If yes, stop now!
        if( ncpeek()->string eq "}" )
        {
            last;
        }

        # Check if next child is "..."
        if( ncpeek()->string eq "..." )
        {
            getNoncommentToken();
            $hasDotDotDot = TRUE;
            last;
        }

        # check for pre-child comment
        my $comment = FALSE;
        if( ASN1tokenizer::peek()->{type} == ASN1Token::BLOCK_COMMENT )
        {
            $comment = ASN1tokenizer::getToken()->string;
        }

        my $specname = getNoncommentToken()->{string};
        my $uniquename = $specname;
        my $childtype = getNoncommentToken()->{string};
        if( ($childtype eq "BIT") || ($childtype eq "OCTET") )
        {
            matchOrFail("STRING");
            $childtype .= " STRING";
        }
        elsif( ($childtype eq "SEQUENCE") && (ncpeek()->string eq "(") )
        {
            $childtype = "SEQ-OF";
        }


        if( isAsn1CoreType($childtype) )
        {
            $uniquename = $nameprefix.$specname;
            processImplicitStruct($uniquename,$specname,$childtype);
        } 

        my $peek = ncpeek()->{string};
        if( $peek eq "OPTIONAL" )
        {
            matchOrFail("OPTIONAL");
            push @opt, TRUE;
            push @isdef, FALSE;
            push @def, FALSE;
        }
        elsif( $peek eq "DEFAULT" )
        {
            matchOrFail("DEFAULT");
            push @isdef, TRUE;
            push @def, getNoncommentToken()->string;
            push @opt, FALSE;
        }
        else
        {
            push @opt, FALSE;
            push @isdef, FALSE;
            push @def, FALSE;
        }

        push @name, $specname;
        push @uname, $uniquename;
        push @ctype, $childtype;
        push @comments, $comment;

        my $more_children_are_ahead = FALSE;
        if( ncpeek()->string eq "," )
        {
            $more_children_are_ahead = TRUE;
            getNoncommentToken(); # consume that comma
        }

        if( ASN1tokenizer::peek()->type == ASN1Token::EOL_COMMENT )
        {
            push @eolcomments, ASN1tokenizer::getToken()->string;
        }
        else
        {
            push @eolcomments, FALSE;
        }

        if( $more_children_are_ahead )
        {
            next;
        }


        #sanity check
        if( ncpeek()->string ne "}" )
        {
            my $foo = ncpeek()->string;
            die "Expected '}' after SEQUENCE child $uniquename (instead got $foo)";
        }
        
        last; #aka break
    }

    my $tail_comment = undef;
    if(ASN1tokenizer::peek()->type == ASN1Token::BLOCK_COMMENT)
    {
        $tail_comment = grabToken()->string;
    }

    # There's only one possible token at this point, right?
    matchOrFail( "}", "SEQ $nameprefix");


    my $seqcont = new SequenceContent;
    for(0..$#name)
    {
        $seqcont->addChild($name[$_],$uname[$_],$ctype[$_],$opt[$_],$comments[$_],$isdef[$_],$def[$_],$eolcomments[$_]);
    }
    if( $tail_comment )
    {
        $seqcont->setTailComment($tail_comment);
    }
    if( $hasDotDotDot )
    {
        $seqcont->setHasDotDotDot(TRUE);
    }

    return $seqcont;
}

sub processChoice($)
{
    my $nameprefix = $_[0];

    # Parsing starts on the next token after the word CHOICE,
    # (which we expect is going to be a "{").
    matchOrFail("{");
    
    my (@name, @uname, @ctype, @comments, @eolcomments);
    my $hasDotDotDot = FALSE;

    while(1)
    {
        # check for pre-child comment
        my $comment = FALSE;
        if( ASN1tokenizer::peek()->{type} == ASN1Token::BLOCK_COMMENT )
        {
            $comment = ASN1tokenizer::getToken()->string;
        }
        
        # Check if next child is "..."
        if( ncpeek()->string eq "..." )
        {
            getNoncommentToken();
            $hasDotDotDot = TRUE;
            last;
        }

        my $specname = getNoncommentToken()->{string};
        my $uniquename = $specname;
        my $childtype = getNoncommentToken()->{string};
        if( ($childtype eq "BIT") || ($childtype eq "OCTET") )
        {
            matchOrFail("STRING");
            $childtype .= " STRING";
        }
        elsif( ($childtype eq "SEQUENCE") && (ncpeek()->string eq "(") )
        {
            $childtype = "SEQ-OF";
        }


        if( isAsn1CoreType($childtype) )
        {
            $uniquename = $nameprefix.$specname;
            processImplicitStruct($uniquename,$specname,$childtype); 
        }

        push @name, $specname;
        push @uname, $uniquename;
        push @ctype, $childtype;
        push @comments, $comment;

        my $more_children_are_ahead = FALSE;
        if( ncpeek()->string eq "," )
        {
            $more_children_are_ahead = TRUE;
            getNoncommentToken(); # consume that comma
        }

        if( ASN1tokenizer::peek()->type == ASN1Token::EOL_COMMENT )
        {
            push @eolcomments, ASN1tokenizer::getToken()->string;
        }
        else
        {
            push @eolcomments, FALSE;
        }

        if( $more_children_are_ahead )
        {
            next;
        }


        #sanity check
        if( ncpeek()->string ne "}" )
        {
            die "Expected '}' after CHOICE $uniquename";
        }
        
        last; #aka break
    }

    my $tail_comment = undef;
    if(ASN1tokenizer::peek()->type == ASN1Token::BLOCK_COMMENT)
    {
        $tail_comment = grabToken()->string;
    }

    # There's only one possible token at this point, right?
    matchOrFail("}", "CHOICE $nameprefix");


    my $chcont = new ChoiceContent;
    for(0..$#name)
    {
        $chcont->addChoice($name[$_],$uname[$_],$ctype[$_],$comments[$_],$eolcomments[$_]);
    }
    if( $tail_comment )
    {
        $chcont->setTailComment($tail_comment);
    }
    if( $hasDotDotDot )
    {
        $chcont->setHasDotDotDot(TRUE);
    }

    return $chcont;
}

sub processSeqOf($)
{
    # Forms:
    # SEQUENCE ( size ) OF sometype
    # where "size" is one of the usual SIZE(n) or (n..n) strings
    # where "sometype" can be a complex type
    my $nameprefix = $_[0];

    # first comes the size
    matchOrFail("(");
    matchOrFail("SIZE");
    my $bounds = processSize();
    matchOrFail(")");
    
    matchOrFail("OF");

    my $comment_after_of = undef;
    if(ASN1tokenizer::peek()->type == ASN1Token::BLOCK_COMMENT)
    {
        $comment_after_of = grabToken()->string;
    }

    my $childtype = getNoncommentToken()->string;
    if( ($childtype eq "BIT") || ($childtype eq "OCTET") )
    {
        matchOrFail("STRING");
        $childtype .= " STRING";
    }
    elsif( ($childtype eq "SEQUENCE") && (ncpeek()->string eq "(") )
    {
        $childtype = "SEQ-OF";
    }

    my $childuname = $nameprefix."seqof";
    if( isAsn1CoreType($childtype) )
    {
        processImplicitStruct($childuname,"seqof",$childtype);
    }
    else
    {
        $childuname = $childtype;
    }

    if( defined($comment_after_of) )
    {
        return [ $childuname, $bounds->[0], $bounds->[1], $comment_after_of ];
    }
    else
    {
        return [ $childuname, $bounds->[0], $bounds->[1] ];
    }
}



# INTEGER (#..#)   <-- note that # might be a symbol; default part is optional
# INTEGER - if type not followed by "(", then we just assume it's the "naked" type 
# above case and _only_ the above case is handled
sub processInteger()
{
    if( ncpeek()->string ne "(" )
    {
        return [];
    }

    matchOrFail("(");
    my $lb = getNoncommentToken()->{string};
    matchOrFail("..");
    my $ub = getNoncommentToken()->{string};
    matchOrFail(")");

    return [ $lb, $ub ];
}

sub processConstantInteger()
{
    my $c = getNoncommentToken()->string;

    my $eol_comment = undef;
    if(ASN1tokenizer::peek()->type eq ASN1Token::EOL_COMMENT)
    {
        $eol_comment = ASN1tokenizer::getToken()->string;
    }
    
    if( defined($eol_comment) )
    {
        return [ $c, $eol_comment ];
    }
    else
    {
        return [ $c ];
    }
}

# this only covers unconstrained, (#), and (#..#)
sub processBitOrOctetString()
{
    # Parsing starts right after "STRING".
    # Possible followers are:
    # 1) (SIZE (n))
    # 2) (SIZE (m..n))
    # Also these, but only for bitstrings (not that it matters)
    # 3) [nothing, aka unconstrained bitstring]
    # 4) (CONTAINING sometype)
    # 5) { someie(1), someie(2), someie(3), ..., someie(n) } (SIZE(n))

    
    # ============
    # DO NOT CONSUME HERE.  If it's type 3, then nothing will be consumed!
    # ============

    # types 1, 2, 4
    if(ncpeek()->string eq "(")
    {
        my $bounds;
        matchOrFail("(");
        
        # type 4
        if( ncpeek()->string eq "CONTAINING" )
        {
            matchOrFail("CONTAINING");
            $bounds = ["--CONTAINING", getNoncommentToken()->{string}];
        }
        # type 1 or 2
        else
        {
            matchOrFail("SIZE");
            $bounds = processSize();
        }

        matchOrFail(")");
        return $bounds;
    }

    # type 5
    elsif(ncpeek()->string eq "{")
    {
        my @fields;
        matchOrFail("{");

        my $comment = undef;
        if(ASN1tokenizer::peek()->type == ASN1Token::BLOCK_COMMENT)
        {
            $comment = grabToken()->string;
        }
        
        # fill @fields with the field names
        while(1)
        {
            push @fields, getNoncommentToken()->string;
            matchOrFail("(");
            if( ncpeek()->string =~ /^[0-9]+$/ ) #sanity check
            {
                getNoncommentToken(); #consume
            }
            else
            {
                die "Oops! Found ".getNoncommentToken()->string." where integer was expected in a bit field element.";
            }
            matchOrFail(")");

            if( ncpeek()->string eq "," )
            {
                getNoncommentToken(); #consume
            }
            else
            {
                last;
            }
        }

        matchOrFail("}");
        matchOrFail("(");
        matchOrFail("SIZE");
        my $size = processSize();
        matchOrFail(")");

        # sanity check
        if($size->[0] != $size->[1])
        {
            die "Oops!  Found a bit field with a variable length \"SIZE\" which I'm not coded to handle.";
        }
        
        if(defined($comment))  { return ["--BITFIELD", $comment, $size->[1], @fields]; }
        else                   { return ["--BITFIELD", $size->[1], @fields]; }
    }

    # type 3 (unconstrained bitstring)
    else
    {
        return -1;
    }

    die("Should be impossible to get here!  End of processString ");
}

# handles tokens immediately following "SIZE" - currently only (#) or (#..#)
sub processSize()
{
    matchOrFail("(");
    my $lb = getNoncommentToken()->{string};

    my $nextToken = getNoncommentToken();
    if( ")" eq $nextToken->{string} )
    {
        return [ $lb, $lb ];
    }
    elsif( ".." eq $nextToken->{string} )
    {
        my $ub = getNoncommentToken()->{string};
        matchOrFail(")");
        return [ $lb, $ub ];
    }
    else
    {
        die("Token ".$nextToken->{string}." from input line ".$nextToken->line." didn't match ')' or '..' as expected in a SIZE construct.");
    }
}

sub processBoolean()
{
    #Processing starts immediately after BOOLEAN
    #Hey, there's nothing to do!  Awesome!
    return undef;
}

sub processNull()
{
    #Processing starts immediately after NULL
    #Hey, there's nothing to do!  Awesome!
    return undef;
}

sub processEnum()
{
    my @a;
    
    #Processing starts immediately after ENUMERATED
    matchOrFail("{","processEnum");

    #now just a comma-separated list of values (with comments possibly mixed in)
    while(1)
    {
        my $t = grabToken();

        if($t->type == ASN1Token::BLOCK_COMMENT)
        {
            push @a, $t->string;
            $t = getNoncommentToken();
        }

        push @a, $t->string;

        if(ncpeek()->string ne "}")
        {
            matchOrFail(",","processEnum");
        }
        else
        {
            if( ASN1tokenizer::peek()->type == ASN1Token::BLOCK_COMMENT )
            {
                push @a, grabToken()->string;
            }
            last;
        }
    }

    matchOrFail("}","processEnum");
    return [@a];
}
















1;
