#class SequenceContent
package SequenceContent;

use strict;
use BooleanTF;


#constructor
sub new {
    my $invocant = shift; #the first arg is the class
    my $class = ref($invocant) || $invocant;
    my $self = {
        num_of_opts => 0,  # how may optional children
        childnames => [],  # array of child names
        childunames => [], # array of unique names
        childtypes => [],  # array of child types
        child_opts => [],  # array of bools, true if child is optional
        block_comments => [],  # array of comments which precede children
        child_isdef => [], # array of bools
        child_defs => [],  # array of default values, FALSE if none
        eol_comments => [],  # array of eol comments after children

        tail_comment => undef,
        has_dotdotdot => FALSE
        # no values taken from the constructor!
    };
    bless $self, $class;
    return $self;
}

sub addChild($$$$$$$)
{
    my ($ob,$cname,$cuname,$ctype,$copt,$ccom,$cisdef,$cdef,$ceol) = @_;

    push @{$ob->{childnames}}, $cname;
    push @{$ob->{childunames}},$cuname;
    push @{$ob->{childtypes}}, $ctype;
    push @{$ob->{child_opts}}, $copt;
    push @{$ob->{block_comments}}, $ccom;
    push @{$ob->{child_isdef}}, $cisdef;
    push @{$ob->{child_defs}}, $cdef;
    push @{$ob->{eol_comments}}, $ceol;

    if($copt)
    {
        $ob->{num_of_opts}++;
    }
}

# =========
# accessors
# =========
sub getNumOpts { return $_[0]->{num_of_opts}; } 

sub getChildArray($)
{
    my ($ob,$x) = @_;

    if( ($x > $#{$ob->{childnames}}) || ($x < 0) )
    {
        return FALSE;
    }

    return [ ${$ob->{childnames}}[$x], 
             ${$ob->{childunames}}[$x],
             ${$ob->{childtypes}}[$x],
             ${$ob->{child_opts}}[$x],
             ${$ob->{block_comments}}[$x],
             ${$ob->{child_isdef}}[$x],
             ${$ob->{child_defs}}[$x],
             ${$ob->{eol_comments}}[$x]     ];
}

sub getNumChildren()
{
    return 1+$#{$_[0]->{childnames}}; 
}

#only useful for pretty printing
sub getMaxNameWidth()
{
    my $maxl = 0;
    foreach( @{$_[0]->{childnames}} )
    {
        if( $maxl < length ) { $maxl = length; }
    }
    return $maxl;
}


sub setTailComment($)
{
    my ($ob,$c) = @_;
    $ob->{tail_comment} = $c;
}

sub getTailComment()
{
    if( defined($_[0]->{tail_comment}) )
    {
        return $_[0]->{tail_comment};
    }

    return FALSE;
}


sub setHasDotDotDot($)
{
    my ($ob,$c) = @_;
    $ob->{has_dotdotdot} = $c;
}

sub getHasDotDotDot()
{
    return $_[0]->{has_dotdotdot};
}

1;
